import copy
import pandas as pd

from bigquery.core.mapping_type import mapping_types


def get_columns_type(_dbstream, schema_name, table_name):
    d = {}
    r = _dbstream.get_data_type(table_name, schema_name)
    for i in r:
        d[i["column_name"]] = i["data_type"]
    return d


def change_type(_dbstream, table_name, column_name, type):
    query = """
    CREATE OR REPLACE TABLE %(table_name)s AS 
    SELECT * EXCEPT (%(column_name)s), CAST(%(column_name)s AS %(type)s) AS %(column_name)s FROM %(table_name)s;
    """ % {
        "table_name": table_name,
        "column_name": column_name,
        "type": type
    }
    _dbstream.execute_query(query)
    return query


def bool_to_str(_dbstream, table_name, column_name):
    query = """
        CREATE OR REPLACE TABLE %(table_name)s AS 
        SELECT * EXCEPT (%(column_name)s), CAST(%(column_name)s AS STRING) AS %(column_name)s FROM %(table_name)s;
        """ % {
        "table_name": table_name,
        "column_name": column_name,
    }
    _dbstream.execute_query(query)
    return query


def change_columns_type(_dbstream, data, other_table_to_update):
    table_name = data["table_name"].split('.')
    columns_type = get_columns_type(_dbstream, table_name=table_name[1], schema_name=table_name[0])
    rows = data["rows"]
    columns_name = data["columns_name"]
    df = pd.DataFrame(rows, columns=columns_name)

    for c in columns_name:
        example = find_sample_value(df, c, columns_name.index(c))[0]
        if isinstance(example, float):
            if columns_type.get(c) != "FLOAT64" and columns_type.get(c) != "STRING":
                change_type(_dbstream, table_name=data["table_name"], column_name=c, type="FLOAT64")
                if other_table_to_update:
                    change_type(_dbstream, table_name=other_table_to_update, column_name=c, type="FLOAT64")
        if isinstance(example, str):
            if columns_type.get(c) != "STRING" and columns_type.get(c) != "BOOL" and columns_type.get(c) != "TIMESTAMP":
                change_type(_dbstream, table_name=data["table_name"], column_name=c, type="STRING")
                if other_table_to_update:
                    change_type(_dbstream, table_name=other_table_to_update, column_name=c, type="STRING")


def columns_type_bool_to_str(_dbstream, data, other_table_to_update):
    table_name = data["table_name"].split('.')
    columns_type = get_columns_type(_dbstream, table_name=table_name[1], schema_name=table_name[0])
    rows = data["rows"]
    columns_name = data["columns_name"]
    df = pd.DataFrame(rows, columns=columns_name)

    for c in columns_name:
        example = find_sample_value(df, c, columns_name.index(c))[0]
        if isinstance(example, str):
            if columns_type.get(c) == "BOOL":
                bool_to_str(_dbstream, table_name=data["table_name"], column_name=c)
                if other_table_to_update:
                    bool_to_str(_dbstream, table_name=other_table_to_update, column_name=c)


def detect_type(_dbstream, name, example, types):
    print('Define type of %s...' % name)
    if types and types.get(name):
        try:
            return mapping_types[types.get(name)]
        except KeyError:
            raise Exception(
                "Type %s missing in mapping_types dictionary [dbstream types] <> [big query types]" % types.get(name)
            )
    try:
        query = "SELECT CAST('%s' as DATE)" % example
        _dbstream.execute_query(query)
        return "DATE"
    except:
        try:
            query = "SELECT CAST('%s' as TIMESTAMP)" % example
            _dbstream.execute_query(query)
            return "TIMESTAMP"
        except:
            pass
    if isinstance(example, bool):
        return "BOOL"
    elif isinstance(example, int):
        return "INT64"
    elif isinstance(example, float):
        return "FLOAT64"
    else:
        return "STRING"


def convert_to_bool(x):
    if x.lower() == "true" or x == 1 or x.lower() == "t":
        return True
    if x.lower() == "false" or x == 0 or x.lower() == "f":
        return False
    else:
        raise Exception


def convert_to_int(x):
    if x[-2:] == ".0":
        return int(x.replace(".0", ""))
    else:
        return int(x)


def len_or_max(s):
    if isinstance(s, str):
        return len(s)
    return s


def find_sample_value(df, name, i):
    df1 = df[name].dropna()
    try:
        df1 = df1.apply(lambda x: str(x))
    except:
        pass
    try:
        df1 = df1.apply(lambda x: convert_to_bool(x))
    except:
        try:
            df1 = df1.apply(lambda x: convert_to_int(x))
        except:
            try:
                df1 = df1.apply(lambda x: float(x))
            except:
                pass
    df1_copy = copy.deepcopy(df1)
    if df1.dtype == 'object':
        df1 = df1.apply(lambda x: (str(x.encode()) if isinstance(x, str) else x) if x is not None else '')
        if df1.empty:
            return None, None
        else:
            return df1_copy[df1.map(len_or_max) == df1.map(len_or_max).max()].iloc[0], \
                   df1_copy[df1.map(len_or_max) == df1.map(len_or_max).min()].iloc[0]
    elif df1.dtype == 'int64':
        max = int(df1.max())
        min = int(df1.min())
        return max, min
    elif df1.dtype == 'float64':
        max = float(df1.max())
        min = float(df1.min())
        return max, min
    else:
        rows = df.values.tolist()
        for row in rows:
            if row[i] is not None:
                return row[i], row[i]
        return None, None



import pandas as pd

from bigquery.core.Column import detect_type, find_sample_value


def get_table_info(_dbstream, table_and_schema_name):
    split = table_and_schema_name.split(".")
    if len(split) == 2:
        table_name = split[1]
        schema_name = split[0]
    else:
        raise Exception("Invalid table or schema name")
    query = "SELECT column_name, data_type, is_nullable FROM %s.INFORMATION_SCHEMA.COLUMNS WHERE TABLE_NAME='%s'" % (
        schema_name, table_name)
    return _dbstream.execute_query(query)


def format_create_table(_dbstream, data):
    columns_name = data["columns_name"]
    rows = data["rows"]
    params = {}
    df = pd.DataFrame(rows, columns=columns_name)
    df = df.where((pd.notnull(df)), None)
    for i in range(len(columns_name)):
        name = columns_name[i]
        example_max, example_min = find_sample_value(df, name, i)
        col = dict()
        col["example"] = example_max
        type_max = detect_type(_dbstream, name=name, example=example_max, types=data.get("types"))
        if type_max == "TIMESTAMP":
            type_min = detect_type(_dbstream, name=name, example=example_min, types=data.get("types"))
            if type_min == type_max:
                col["type"] = type_max
            else:
                col["type"] = type_min
        else:
            col["type"] = type_max
        params[name] = col

    query = """"""
    query = query + "CREATE TABLE %(table_name)s ("
    col = list(params.keys())
    for i in range(len(col)):
        k = col[i]
        string_example = " --example:" + str(params[k]["example"])[:10].replace("\n", "").replace("%", "") + ''
        if i == len(col) - 1:
            query = query + "\n     `" + k + '` ' + params[k]["type"] + string_example
        else:
            query = query + "\n     `" + k + '` ' + params[k]["type"] + ',' + string_example
    query = query + "\n )"
    print(query)
    return query


def create_table(_dbstream, data, other_table_to_update):
    query = format_create_table(_dbstream, data)
    try:
        _dbstream.execute_query(query % {"table_name": data["table_name"]})
        if other_table_to_update:
            _dbstream.execute_query(query % {"table_name": other_table_to_update})
    except Exception as e:
        if " was not found " in str(e).lower() and " dataset " in str(e).lower():
            schema_name = data['table_name'].split(".")[0]
            _dbstream.create_schema(schema_name)
        else:
            raise e


def create_columns(_dbstream, data, other_table_to_update):
    table_name = data["table_name"]
    rows = data["rows"]
    columns_name = data["columns_name"]
    infos = get_table_info(_dbstream, table_name)
    all_column_in_table = [e['column_name'] for e in infos]
    df = pd.DataFrame(rows, columns=columns_name)
    df = df.where((pd.notnull(df)), None)
    query_table = "alter table %s \n" % table_name
    if other_table_to_update:
        query_other_table = "alter table %s \n" % other_table_to_update
    queries_table = []
    queries_other_table = []
    for column_name in columns_name:
        if column_name not in all_column_in_table:
            example_max, example_min = find_sample_value(df, column_name, columns_name.index(column_name))
            type_max = detect_type(_dbstream, name=column_name, example=example_max, types=data.get("types"))
            if type_max == "TIMESTAMP":
                type_min = detect_type(_dbstream, name=column_name, example=example_min, types=data.get("types"))
                if type_min == type_max:
                    type_ = type_max
                else:
                    type_ = "STRING"
            else:
                type_ = type_max
            query = """
            add COLUMN %s %s
            """ % (column_name, type_)
            queries_table.append(query)
            if other_table_to_update:
                query = """add COLUMN %s %s""" % (column_name, type_)
                queries_other_table.append(query)
    if queries_table:
        query = query_table + ',\n '.join(queries_table) + ";"
        _dbstream.execute_query(query)
    if queries_other_table:
        query = query_other_table + ',\n '.join(queries_other_table) + ";"
        _dbstream.execute_query(query)
    return 0



import copy
import os
import random
import re
import dbstream
import time
import google.cloud.bigquery
import pandas as pd
from google.cloud import bigquery
from google.cloud.bigquery import SchemaField
from google.cloud.bigquery.dbapi import Cursor
from googleauthentication import GoogleAuthentication
from bigquery.core.Column import change_columns_type, columns_type_bool_to_str, \
    find_sample_value, detect_type, get_columns_type
from bigquery.core.tools.print_colors import C
from bigquery.core.Table import create_table, create_columns
from google.cloud.bigquery.job import QueryJobConfig, ScriptOptions


class BigQueryDBStream(dbstream.DBStream):
    def __init__(self,
                 instance_name,
                 client_id,
                 google_auth: GoogleAuthentication,
                 dataset_location=None,
                 tmp_folder_path="./"):
        super().__init__(instance_name, client_id=client_id)
        self.instance_type_prefix = "BIGQ"
        self.google_auth = google_auth
        self.ssh_init_port = 6543
        self.dataset_location = dataset_location
        self.tmp_folder_path = tmp_folder_path

    def connection(self):
        try:
            con = google.cloud.bigquery.client.Client(
                project=os.environ["BIG_QUERY_PROJECT_ID"],
                credentials=self.google_auth.credentials()
            )
        except google.cloud.bigquery.dbapi.OperationalError:
            time.sleep(2)
            if self.ssh_tunnel:
                self.ssh_tunnel.close()
                self.create_tunnel()
            con = google.cloud.bigquery.client.Client(
                project=os.environ["BIG_QUERY_PROJECT_ID"],
                credentials=self.google_auth.credentials()
            )
        return con

    def _execute_query_custom(self, query):
        client = self.connection()
        con = google.cloud.bigquery.dbapi.connect(client=client)
        cursor = Cursor(con)
        try:
            cursor.execute(query, job_config=QueryJobConfig(script_options=ScriptOptions(statement_timeout_ms=120000)))
        except Exception as e:
            cursor.close()
            con.close()
            raise e
        con.commit()
        try:
            result = cursor.fetchall()
        except:
            result = None
        cursor.close()
        con.close()
        query_create_table = re.search("(?i)(?<=((create table ))).*(?= as)", query)
        if result:
            return [dict(r) for r in result]
        elif query_create_table:
            return {'execute_query': query_create_table}
        else:
            empty_list = []
            return empty_list

    def _send(self, data, replace=True, batch_size=1000):
        print(C.WARNING + "Initiate send_to_bigquery on table " + data["table_name"] + "..." + C.ENDC)

        if replace:
            print(C.WARNING + "Table will be cleaned: " + data["table_name"] + C.ENDC)

        total_rows = len(data["rows"])

        # Construct a BigQuery client object.
        client = self.connection()
        columns_name = data["columns_name"]

        df = pd.DataFrame(data["rows"], columns=columns_name)

        randint = str(random.randrange(1000000))
        tmp_csv_path = self.tmp_folder_path + randint + "%s.csv" % data["table_name"].replace('.', '_')
        df.to_csv(tmp_csv_path, index=False)

        params = {}
        df = df.where((pd.notnull(df)), None)

        table_name = data["table_name"].split('.')
        columns_type = get_columns_type(self, table_name=table_name[1], schema_name=table_name[0])

        for i in range(len(columns_name)):
            name = columns_name[i]
            col = dict()
            if name in columns_type.keys():
                col["type"] = columns_type[name]
                params[name] = col
                continue
            example_max, example_min = find_sample_value(df, name, i)
            col = dict()
            col["example"] = example_max
            type_max = detect_type(self, name=name, example=example_max, types=data.get("types"))
            if type_max == "TIMESTAMP" or type_max == "DATE":
                type_min = detect_type(self, name=name, example=example_min, types=data.get("types"))
                if type_min == type_max:
                    col["type"] = type_max
                else:
                    col["type"] = type_min
            else:
                col["type"] = type_max
            params[name] = col

        schema = [
            SchemaField(name=c, field_type=params[c]["type"]) for c in params.keys()
        ]

        job_config = bigquery.LoadJobConfig(
            source_format=bigquery.SourceFormat.CSV,
            skip_leading_rows=1,
            schema=schema,
            autodetect=True,
            allow_quoted_newlines=True,
            write_disposition="WRITE_TRUNCATE" if replace else "WRITE_APPEND"
        )

        table_id = os.environ["BIG_QUERY_PROJECT_ID"] + "." + data["table_name"]

        with open(tmp_csv_path, "rb") as source_file:
            job = client.load_table_from_file(source_file, table_id, job_config=job_config)

        job.result()  # Waits for the job to complete.

        table = client.get_table(table_id)  # Make an API request.
        print(
            "Loaded {} rows and {} columns to {}".format(
                table.num_rows, len(table.schema), table_id
            )
        )

        print(C.HEADER + str(total_rows) + ' rows sent to BigQuery table ' + data["table_name"] + C.ENDC)
        print(C.OKGREEN + "[OK] Sent to bigquery" + C.ENDC)
        os.remove(tmp_csv_path)
        print(C.OKGREEN + "[OK] " + tmp_csv_path + " deleted" + C.ENDC)
        return 0

    def _send_data_custom(self,
                          data,
                          replace=True,
                          batch_size=1000,
                          other_table_to_update=None,
                          n=1
                          ):
        """
        data = {
            "table_name" 	: 'name_of_the_redshift_schema' + '.' + 'name_of_the_redshift_table' #Must already exist,
            "columns_name" 	: [first_column_name,second_column_name,...,last_column_name],
            "rows"		: [[first_raw_value,second_raw_value,...,last_raw_value],...]
        }
        """
        data_copy = copy.deepcopy(data)
        try:
            self._send(data, replace=replace, batch_size=batch_size)
        except Exception as e:
            error_lowercase = str(e).lower()
            if ("could not parse " in error_lowercase and "int64" in error_lowercase) \
                    or ("could not parse " in error_lowercase and "double" in error_lowercase):
                change_columns_type(
                    self,
                    data=data_copy,
                    other_table_to_update=other_table_to_update
                )

            elif "could not parse " in error_lowercase and "bool" in error_lowercase:
                columns_type_bool_to_str(
                    self,
                    data=data_copy,
                    other_table_to_update=other_table_to_update
                )
            elif " not found:" in error_lowercase and (
                    " table " in error_lowercase or " dataset " in error_lowercase):
                print("Destination table doesn't exist! Will be created")
                create_table(
                    self,
                    data=data_copy,
                    other_table_to_update=other_table_to_update
                )
            elif " provided schema does not match table " in error_lowercase and " cannot add fields " in error_lowercase:
                create_columns(
                    self,
                    data=data_copy,
                    other_table_to_update=other_table_to_update
                )
            else:
                raise e

            self._send_data_custom(data_copy, replace=replace, batch_size=batch_size,
                                   other_table_to_update=other_table_to_update, n=n)

    def clean(self, selecting_id, schema_prefix, table):
        print('trying to clean table %s.%s using %s' % (schema_prefix, table, selecting_id))
        cleaning_query = """
                DELETE FROM %(schema_name)s.%(table_name)s WHERE %(id)s IN (SELECT distinct %(id)s FROM %(schema_name)s.%(table_name)s_temp);
                INSERT INTO %(schema_name)s.%(table_name)s 
                SELECT * FROM %(schema_name)s.%(table_name)s_temp;
                DELETE FROM %(schema_name)s.%(table_name)s_temp WHERE 1=1;
                """ % {"table_name": table,
                       "schema_name": schema_prefix,
                       "id": selecting_id}
        self.execute_query(cleaning_query)
        print('cleaned')

    def get_max(self, schema, table, field, filter_clause=""):
        try:
            print("SELECT max(%s) as max FROM %s.%s %s" % (field, schema, table, filter_clause))
            r = self.execute_query("SELECT max(%s) as max FROM %s.%s %s" % (field, schema, table, filter_clause))
            return r[0]["max"]
        except Exception as e:
            if "was not found" in str(e):
                return None
            raise e

    def get_data_type(self, table_name, schema_name):
        query = """ SELECT column_name, data_type FROM %s.INFORMATION_SCHEMA.COLUMNS WHERE TABLE_NAME='%s' """ \
                % (schema_name, table_name)
        return self.execute_query(query=query)

    def create_view_from_columns(self, view_name, columns, schema_name, table_name):
        view_query = '''DROP VIEW IF EXISTS %s.%s ;CREATE VIEW %s.%s as (SELECT %s FROM %s.%s)''' \
                     % (schema_name, view_name, schema_name, view_name, columns, schema_name, table_name)
        self.execute_query(view_query)

    def create_schema(self, schema_name):
        con = self.connection()
        dataset = google.cloud.bigquery.Dataset(con.project + "." + schema_name)
        if self.dataset_location:
            dataset.location = self.dataset_location
        con.create_dataset(dataset)

    def drop_schema(self, schema_name):
        con = self.connection()
        con.delete_dataset(dataset=con.project + "." + schema_name, delete_contents=True, not_found_ok=True)



from setuptools import setup, find_packages

with open('README.rst') as f:
    readme = f.read()

setup(
    name='bigquery',
    version='0.0.17',
    description='Easily send data to Big Query',
    long_description=readme,
    author='Dacker',
    author_email='hello@dacker.co',
    url='https://github.com/dacker-team/bigquery',
    keywords='send data bigquery easy',
    packages=find_packages(exclude=('tests', 'docs')),
    python_requires='>=3',
    install_requires=[
        "dbstream>=0.0.19",
        "google-cloud-bigquery>=2.4.0",
        "googleauthentication>=0.0.12",
        "google-cloud-bigquery-storage>=2.1.0",
        "pandas>=1.1.4",
        "pyarrow>=2.0.0"
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)
